package hxie.framework.webservice.jbossws.handler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Node;

public class WSSEUsernameTokenHandler implements
		SOAPHandler<SOAPMessageContext> {
	private static class UserNameTokenNamespaceContext implements
			NamespaceContext {
		private static final Map<String, Collection<String>> NameSpaceURIToPrefixMap = createNameSpaceURIToPrefixMap();
		private static final Map<String, String> PrefixToNameSpaceURIMap = createPrefixToNameSpaceURIMap();

		private static Map<String, Collection<String>> createNameSpaceURIToPrefixMap() {
			HashMap<String, Collection<String>> map = new HashMap<String, Collection<String>>();
			Collection<String> prefixes;

			prefixes = new ArrayList<String>(1);
			prefixes.add(WSSE_PREFIX);
			map.put(WSSE_NAMESPACE_URI, prefixes);

			prefixes = new ArrayList<String>(1);
			prefixes.add(WSU_PREFIX);
			map.put(WSU_NAMESPACE_URI, prefixes);

			prefixes = new ArrayList<String>(1);
			prefixes.add(SOAP_ENVELOPE_PREFIX);
			map.put(SOAP_ENVELOPE_URI, prefixes);

			prefixes = new ArrayList<String>(1);
			prefixes.add(XMLConstants.DEFAULT_NS_PREFIX);
			map.put(XMLConstants.NULL_NS_URI, prefixes);

			prefixes = new ArrayList<String>(1);
			prefixes.add(XMLConstants.XML_NS_PREFIX);
			map.put(XMLConstants.XML_NS_URI, prefixes);

			prefixes = new ArrayList<String>(1);
			prefixes.add(XMLConstants.XMLNS_ATTRIBUTE);
			map.put(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefixes);

			return map;
		}

		private static Map<String, String> createPrefixToNameSpaceURIMap() {
			HashMap<String, String> map = new HashMap<String, String>();
			map.put(WSSE_PREFIX, WSSE_NAMESPACE_URI);
			map.put(WSU_PREFIX, WSU_NAMESPACE_URI);
			map.put(SOAP_ENVELOPE_PREFIX, SOAP_ENVELOPE_URI);
			map.put(XMLConstants.DEFAULT_NS_PREFIX, XMLConstants.NULL_NS_URI);
			map.put(XMLConstants.XML_NS_PREFIX, XMLConstants.XML_NS_URI);
			map.put(XMLConstants.XMLNS_ATTRIBUTE,
					XMLConstants.XMLNS_ATTRIBUTE_NS_URI);
			return map;
		}

		@Override
		public String getNamespaceURI(String prefix) {
			if (prefix == null) {
				throw new IllegalArgumentException("Null prefix");
			}
			if (PrefixToNameSpaceURIMap.containsKey(prefix)) {
				return PrefixToNameSpaceURIMap.get(prefix);
			} else {
				return XMLConstants.NULL_NS_URI;
			}
		}

		@Override
		public String getPrefix(String namespaceURI) {
			if (namespaceURI == null) {
				throw new IllegalArgumentException("Null namespaceURI");
			}

			Collection<String> prefixes = NameSpaceURIToPrefixMap
					.get(namespaceURI);
			if (prefixes == null || prefixes.isEmpty()) {
				return null;
			} else {
				return prefixes.iterator().next();
			}
		}

		@SuppressWarnings("unchecked")
		@Override
		public Iterator getPrefixes(String namespaceURI) {
			if (namespaceURI == null) {
				throw new IllegalArgumentException("Null namespaceURI");
			}
			Collection<String> prefixes = NameSpaceURIToPrefixMap
					.get(namespaceURI);
			if (prefixes == null || prefixes.isEmpty()) {
				return Collections.unmodifiableCollection(
						Collections.emptyList()).iterator();
			} else {
				return Collections.unmodifiableCollection(prefixes).iterator();
			}
		}

	}

	public static final String Header_PART = "Header";
	private static Log logger = LogFactory
			.getLog(WSSEUsernameTokenHandler.class);
	public static final String SECURITY_PART = "Security";
	public static final String SOAP_ENVELOPE_PREFIX = "env";
	public static final String SOAP_ENVELOPE_URI = "http://schemas.xmlsoap.org/soap/envelope/";
	public static final String SOAP_HEADER_PART = "Header";
	public static final String USECREATED_PART = "Created";
	public static final String USERNAMETOKEN_PART = "UsernameToken";
	public static final String WSSE_NAMESPACE_URI = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";
	public static final String WSSE_PREFIX = "wsse";
	public static final String WSU_NAMESPACE_URI = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd";
	public static final String WSU_PREFIX = "wsu";

	@Override
	public void close(MessageContext context) {
		// Nothing to do
	}

	@Override
	public Set<QName> getHeaders() {
		HashSet<QName> headers = new HashSet<QName>();
		// Process WS-Security UsernameToken Profile
		headers.add(new QName(WSSE_NAMESPACE_URI, SECURITY_PART,
				XMLConstants.DEFAULT_NS_PREFIX));
		return headers;
	}

	@Override
	public boolean handleFault(SOAPMessageContext context) {
		// Nothing to do
		return true;
	}

	/**
	 * convert wsse:Security/wsse:UsernameToken/wsu:Created to
	 * wsse:Security/wsse:UsernameToken/wsse:Created
	 * 
	 * @param context
	 * @return
	 */
	protected boolean handleInboundMessage(SOAPMessageContext context) {
		try {
			SOAPHeader head =  context.getMessage().getSOAPHeader();
			XPath xPath = XPathFactory.newInstance().newXPath();
			xPath.setNamespaceContext(new UserNameTokenNamespaceContext());

			// xpath = "./wsse:Security/wsse:UsernameToken/wsu:Created";
			String useCreatedExpr = String.format(
					"./%1$s:%2$s/%1$s:%3$s/%4$s:%5$s", WSSE_PREFIX,
					SECURITY_PART, USERNAMETOKEN_PART, WSU_PREFIX,
					USECREATED_PART);

			Node useCreated = (Node) xPath.evaluate(useCreatedExpr, head,
					XPathConstants.NODE);
			if (useCreated != null && useCreated instanceof SOAPElement) {
				// standard wsu:Created found, which is going to be converted to
				// wsse:Create
				logger
						.warn("Change wsu:Created -> wsse:Created, see https://jira.jboss.org/jira/browse/JBWS-2640");
				((SOAPElement) useCreated).setElementQName(new QName(
						WSSE_NAMESPACE_URI, USECREATED_PART));

			}
			return true;
		} catch (SOAPException e) {
			logger.error(e);
		} catch (XPathExpressionException e) {
			logger.error(e);
		}
		return false;
	}

	@Override
	public boolean handleMessage(SOAPMessageContext context) {
		Boolean outbound = (Boolean) context
				.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);
		if (outbound == null)
			throw new IllegalStateException("Cannot obtain required property: "
					+ MessageContext.MESSAGE_OUTBOUND_PROPERTY);

		return outbound ? handleOutboundMessage(context)
				: handleInboundMessage(context);
	}

	protected boolean handleOutboundMessage(SOAPMessageContext context) {
		try {
			SOAPHeader head = context.getMessage().getSOAPHeader();
			XPath xPath = XPathFactory.newInstance().newXPath();
			xPath.setNamespaceContext(new UserNameTokenNamespaceContext());
			// xpath ="./wsse:Security/wsse:UsernameToken/wsse:Created";
			String useCreatedExpr = String.format(
					"./%1$s:%2$s/%1$s:%3$s/%1$s:%4$s", WSSE_PREFIX,
					SECURITY_PART, USERNAMETOKEN_PART, USECREATED_PART);
			Node useCreated = (Node) xPath.evaluate(useCreatedExpr, head,
					XPathConstants.NODE);
			if (useCreated != null && useCreated instanceof SOAPElement) {
				// JBossWS generated wsse:Created found, which should be
				// wsu:Create
				logger
						.warn("Change wsse:Created -> wsu:Created, see https://jira.jboss.org/jira/browse/JBWS-2640");
				((SOAPElement) useCreated).setElementQName(new QName(
						WSU_NAMESPACE_URI, USECREATED_PART));
			}
			return true;
		} catch (SOAPException e) {
			logger.error(e);
		} catch (XPathExpressionException e) {
			logger.error(e);
		}
		return false;
	}
}
